import logging
from pprint import pformat
from functools import partial

def get_logger(env):
  logger = logging.getLogger("utils")
  lvl = env.NBOX_LOG_LEVEL("info").upper()
  logger.setLevel(getattr(logging, lvl))

  if env.NBOX_JSON_LOG(False):
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(JsonFormatter(
      '%(timestamp)s %(levelname)s %(message)s %(filename)s %(lineno)d',
      timestamp=True
    ))
    logger.addHandler(logHandler)
  else:
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(InlineFormatter(
      '[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d] %(message)s',
      datefmt = "%Y-%m-%dT%H:%M:%S%z"
    ))
    logger.addHandler(logHandler)

  return logger



##############
# custom logger class for simplicity
class LogObject:
  def __init__(self, log, msg):
    self.log = log
    self.msg = msg

def logobj(msg, **kwargs):
  return LogObject(kwargs, msg)


class InlineFormatter(logging.Formatter):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def format(self, record):
    # print(record.msg, type(record.msg))
    if type(record.msg) == LogObject:
      _l: LogObject = record.msg
      record.msg = f"{_l.msg}\n" + json_dumps_pretty(_l.log)
    # '[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d] %(message)s'
    return super().format(record)


##############

'''
This library is provided to allow standard python logging
to output log data as JSON formatted strings
'''
import re
import json
import logging
import traceback
from inspect import istraceback
from datetime import date, datetime, time, timezone

class JsonFormatter(logging.Formatter):
  """
  A custom formatter to format logging records as json strings.
  Extra values will be formatted as str() if not supported by
  json default encoder
  """
  # skip natural LogRecord attributes
  # http://docs.python.org/library/logging.html#logrecord-attributes
  RESERVED_ATTRS = (
    'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename',
    'funcName', 'levelname', 'levelno', 'lineno', 'module',
    'msecs', 'message', 'msg', 'name', 'pathname', 'process',
    'processName', 'relativeCreated', 'stack_info', 'thread', 'threadName')

  def __init__(self, *args, **kwargs):
    """
    :param json_default: a function for encoding non-standard objects
      as outlined in http://docs.python.org/2/library/json.html
    :param json_encoder: optional custom encoder
    :param json_serializer: a :meth:`json.dumps`-compatible callable
      that will be used to serialize the log record.
    :param json_indent: an optional :meth:`json.dumps`-compatible numeric value
      that will be used to customize the indent of the output json.
    :param prefix: an optional string prefix added at the beginning of
      the formatted string
    :param rename_fields: an optional dict, used to rename field names in the output.
      Rename message to @message: {'message': '@message'}
    :param static_fields: an optional dict, used to add fields with static values to all logs
    :param json_indent: indent parameter for json.dumps
    :param json_ensure_ascii: ensure_ascii parameter for json.dumps
    :param reserved_attrs: an optional list of fields that will be skipped when
      outputting json log record. Defaults to all log record attributes:
      http://docs.python.org/library/logging.html#logrecord-attributes
    :param timestamp: an optional string/boolean field to add a timestamp when
      outputting the json log record. If string is passed, timestamp will be added
      to log record using string as key. If True boolean is passed, timestamp key
      will be "timestamp". Defaults to False/off.
    """
    self.json_serializer = json.dumps
    reserved_attrs = kwargs.pop("reserved_attrs", JsonFormatter.RESERVED_ATTRS)
    self.reserved_attrs = dict(zip(reserved_attrs, reserved_attrs))
    self.timestamp = kwargs.pop("timestamp", False)

    logging.Formatter.__init__(self, *args, **kwargs)

    # Parses format string looking for substitutions
    # This method is responsible for returning a list of fields (as strings) to include in all log messages.
    standard_formatters = re.compile(r'\((.+?)\)', re.IGNORECASE)
    self._required_fields = standard_formatters.findall(self._fmt)
    self._skip_fields = dict(zip(self._required_fields, self._required_fields))
    self._skip_fields.update(self.reserved_attrs)

  def format(self, record):
    """Formats a log record and serializes to json"""

    log_record = {
      "timestamp": datetime.fromtimestamp(record.created, tz=timezone.utc),
      "levelname": record.levelname,
      "logtype": "nbox-json-log-v1",
      "log": {
        "code": {
          "file": record.filename,
          "line": record.lineno,
        },
      }
    }

    if isinstance(record.msg, dict):
      log_record["log"]["data"] = record.msg
    elif isinstance(record.msg, LogObject):
      _lo: LogObject = record.msg
      log_record["log"]["data"] = _lo.log
      log_record["message"] = _lo.msg
    elif record.msg:
      lines = record.getMessage().splitlines()
      log_record["log"]["message"] = lines
      if len(lines[0]) > 100:
        log_record["message"] = lines[0][:100] + "..."
      else:
        log_record["message"] = lines[0]
    else:
      log_record["log"]["message"] = "click to open"

    json_str = json_dumps_tight(log_record)
    return json_str


class _JsonEncoder(json.JSONEncoder):
  """
  A custom encoder extending the default JSONEncoder
  """
  def default(self, obj):
    if isinstance(obj, (date, datetime, time)):
      return self.format_datetime_obj(obj)
    elif istraceback(obj):
      return ''.join(traceback.format_tb(obj)).strip()
    elif type(obj) == Exception \
        or isinstance(obj, Exception) \
        or type(obj) == type:
      return str(obj)
    try:
      return super(_JsonEncoder, self).default(obj)
    except TypeError:
      try:
        return str(obj)
      except Exception:
        return None

  def format_datetime_obj(self, obj):
    return obj.isoformat()

json_dumps_tight = lambda log_record: json.dumps(
  log_record,
  cls = _JsonEncoder,
  ensure_ascii = True,
  separators = (',', ':'),
)

json_dumps_pretty = lambda log_record: json.dumps(
  log_record,
  cls = _JsonEncoder,
  ensure_ascii = True,
  indent = 2,
  separators = (',', ': '),
)